from setuptools import setup, find_packages

setup(name='irbench-vision',
      version='0.1.9',
      description='Image Retrieval Performance Benchmark on Large-scale Dataset',
      url='',
      author='Minchul Shin',
      author_email='min.nashory@navercorp.com',
      license='MIT',
      packages=find_packages(),
      zip_safe=False)
